import logging
from collections import defaultdict
from enum import Enum, auto
from typing import Dict, List

import pygtrie

from spotcrates.common import NotFoundException, BaseLookup

logger = logging.getLogger(__name__)


class InvalidFilterException(Exception):
    pass


class FilterType(Enum):
    CONTAINS = auto()
    EQUALS = auto()
    STARTS = auto()
    ENDS = auto()
    GREATER = auto()
    LESS = auto()
    GREATER_EQUAL = auto()
    LESS_EQUAL = auto()

    def test(self, filter_val, target_val) -> bool:
        if self == FilterType.CONTAINS:
            return str(filter_val).lower() in str(target_val).lower()
        elif self == FilterType.EQUALS:
            return str(filter_val).lower() == str(target_val).lower()
        elif self == FilterType.STARTS:
            return str(target_val).lower().startswith(str(filter_val).lower())
        elif self == FilterType.ENDS:
            return str(target_val).lower().endswith(str(filter_val).lower())
        elif self == FilterType.GREATER:
            return int(target_val) > int(filter_val)
        elif self == FilterType.GREATER_EQUAL:
            return int(target_val) >= int(filter_val)
        elif self == FilterType.LESS:
            return int(target_val) < int(filter_val)
        elif self == FilterType.LESS_EQUAL:
            return int(target_val) <= int(filter_val)
        else:
            raise NotFoundException(f"Unhandled filter type {self}")


class FieldName(Enum):
    SPOTIFY_ID = auto()
    PLAYLIST_NAME = auto()
    SIZE = auto()
    OWNER = auto()
    PLAYLIST_DESCRIPTION = auto()
    ALL = auto()

    @staticmethod
    def list_regular_fields():
        for field in FieldName:
            if field == FieldName.ALL:
                continue
            yield field


class FilterLookup(BaseLookup):

    def eval_filter_type(self, filter_type) -> FilterType:
        filter_type_type = type(filter_type)
        if filter_type_type is FilterType:
            return filter_type
        elif filter_type_type is str:
            return self.find(filter_type)
        else:
            raise InvalidFilterException(f"Invalid filter type {filter_type_type}")

    def _init_lookup(self):
        lookup = pygtrie.CharTrie()
        lookup["c"] = FilterType.CONTAINS
        lookup["eq"] = FilterType.EQUALS
        lookup["s"] = FilterType.STARTS
        lookup["en"] = FilterType.ENDS
        lookup["g"] = FilterType.GREATER
        lookup["l"] = FilterType.LESS
        lookup["ge"] = FilterType.GREATER_EQUAL
        lookup["leq"] = FilterType.LESS_EQUAL

        return lookup


class FieldLookup(BaseLookup):

    def eval_field_name(self, field_name):
        field_name_type = type(field_name)
        if field_name_type is FieldName:
            return field_name
        elif field_name_type is str:
            return self.find(field_name)
        else:
            raise InvalidFilterException(f"Invalid field name type {field_name_type}")

    def _init_lookup(self):
        lookup = pygtrie.CharTrie()
        lookup["n"] = FieldName.PLAYLIST_NAME
        lookup["p"] = FieldName.PLAYLIST_NAME
        lookup["pl"] = FieldName.PLAYLIST_NAME
        lookup["pn"] = FieldName.PLAYLIST_NAME
        lookup["s"] = FieldName.SIZE
        lookup["ps"] = FieldName.SIZE
        lookup["d"] = FieldName.PLAYLIST_DESCRIPTION
        lookup["pd"] = FieldName.PLAYLIST_DESCRIPTION
        lookup["c"] = FieldName.SIZE
        lookup["s"] = FieldName.SIZE
        lookup["o"] = FieldName.OWNER
        lookup["a"] = FieldName.ALL

        return lookup


class FieldFilter:

    def __init__(self, field, filter_type, value):
        """Represents a filter with the given settings.

        :param field: The name of the field to filter.
        :param filter_type: The type of filter to apply.
        :param value: The value to test filter against.
        """
        self.filter_lookup = FilterLookup()
        self.field_lookup = FieldLookup()
        self.filter_type = self.filter_lookup.eval_filter_type(filter_type)
        self.field = self.field_lookup.eval_field_name(field)
        self.value = value

    def passes(self, target_value) -> bool:
        """Returns whether the given value passes the configured filter.

        :param target_value: The value to evaluate.
        :return: Whether the value passes the filter.
        """
        return self.filter_type.test(self.value, target_value)

    def __repr__(self):
        return f"FieldFilter({self.field}, {self.value}, {self.filter_type})"

    def __eq__(self, other):
        if isinstance(other, FieldFilter):
            return (
                    self.field == other.field
                    and self.value == other.value
                    and self.filter_type == other.filter_type
            )
        return NotImplemented


def parse_filters(filters: str) -> Dict[FieldName, List[FieldFilter]]:
    """Returns a dict keyed by field with values being a list of
    name:PoP
    size:gt:22
    name:twin


    :param filters: A str with a comma-separated list of filters
    :return: A map of field names to lower-case "contains" filter strings.
    """
    parsed_filters: Dict[FieldName, List[FieldFilter]] = defaultdict(list)

    if not filters:
        return parsed_filters

    raw_filters = [raw_filter.strip() for raw_filter in filters.split(",")]
    split_raw_filters = [raw_filter.split(":") for raw_filter in raw_filters]

    for raw_exp in split_raw_filters:
        exp_field_count = len(raw_exp)

        if exp_field_count < 1:
            raise InvalidFilterException(
                f"Invalid filter expression {':'.join(raw_exp)}"
            )

        stripped_exp = [field.strip() for field in raw_exp]

        if exp_field_count == 1:
            field_filter = FieldFilter(
                FieldName.ALL, FilterType.CONTAINS, stripped_exp[0]
            )
        elif exp_field_count == 2:
            field_filter = FieldFilter(
                stripped_exp[0], FilterType.CONTAINS, stripped_exp[1]
            )
        else:
            field_filter = FieldFilter(
                stripped_exp[0], stripped_exp[1], stripped_exp[2]
            )
        parsed_filters[field_filter.field].append(field_filter)

    return parsed_filters


def filter_list(items, filters):
    """Evaluates the given list of values against the given list of filters, returning
    items that pass all of the filters.

    :param items: The values to filter.
    :param filters: The filters to apply.
    :return: The values that pass all of the filters.
    """
    parsed_filters = parse_filters(filters)

    filtered_items = items
    for field in FieldName:
        field_filters = parsed_filters[field]
        if field_filters:
            for cur_filter in field_filters:
                matching_items = []
                for cur_item in filtered_items:
                    if FieldName.ALL == field:
                        for regular_field in FieldName.list_regular_fields():
                            item_field = cur_item.get(regular_field)
                            if cur_filter.passes(item_field):
                                matching_items.append(cur_item)
                    else:
                        item_field = cur_item.get(field)
                        if cur_filter.passes(item_field):
                            matching_items.append(cur_item)

                filtered_items = list(
                    {v[FieldName.SPOTIFY_ID]: v for v in matching_items}.values()
                )

    return filtered_items


class SortType(Enum):
    ASCENDING = auto()
    DESCENDING = auto()


class SortLookup(BaseLookup):
    def eval_sort_type(self, sort_type):
        sort_type_type = type(sort_type)
        if sort_type_type is SortType:
            return sort_type
        elif sort_type_type is str:
            return self.find(sort_type)
        else:
            raise InvalidFilterException(f"Invalid sort type {sort_type_type}")

    def _init_lookup(self):
        lookup = pygtrie.CharTrie()
        lookup["a"] = SortType.ASCENDING
        lookup["d"] = SortType.DESCENDING
        lookup["r"] = SortType.DESCENDING
        return lookup


class FieldSort:
    def __init__(self, field, sort_type):
        self.sort_lookup = SortLookup()
        self.field_lookup = FieldLookup()
        self.sort_type = self.sort_lookup.eval_sort_type(sort_type)
        self.field = self.field_lookup.eval_field_name(field)


def parse_sorts(sorts: str) -> List[FieldSort]:
    """Evaluates the sort expressions in the given string, returning the parsed sorts.

    :param sorts: The sorts to parse.
    :return: The parsed sorts.
    """
    parsed_sorts: List[FieldSort] = []

    if not sorts:
        return parsed_sorts

    raw_sorts = [raw_sort.strip() for raw_sort in sorts.split(",")]
    split_raw_sorts = [raw_sort.split(":") for raw_sort in raw_sorts]

    for raw_exp in split_raw_sorts:
        exp_field_count = len(raw_exp)

        stripped_exp = [field.strip() for field in raw_exp]
        if exp_field_count == 1:
            parsed_sorts.append(FieldSort(stripped_exp[0], SortType.ASCENDING))
        elif exp_field_count == 2:
            parsed_sorts.append(FieldSort(stripped_exp[0], stripped_exp[1]))

    return parsed_sorts


def sort_list(items: List, sort_exp: str):
    parsed_sorts = parse_sorts(sort_exp)

    if not parsed_sorts:
        return items

    if len(parsed_sorts) > 1:
        logger.warning("Only the first sort field is currently applied")

    sorter = parsed_sorts[0]

    return sorted(
        items,
        key=lambda d: d[sorter.field],
        reverse=sorter.sort_type == SortType.DESCENDING,
    )



#!/usr/bin/env python

"""
CLI runner for Spotify automation.
"""

import argparse
import logging
import re
import sys
from typing import Dict, Any, List

import pygtrie
import tomli_w

from spotcrates.common import BaseLookup, truncate_long_value, get_spotify_handle, DEFAULT_CONFIG_FILE, get_config, \
    ValueFilter
from spotcrates.filters import FieldName

import importlib.metadata

__version__ = importlib.metadata.version('spotcrates')

__author__ = "cmayes"

from pathlib import Path

from spotcrates.playlists import Playlists, PlaylistResult

# Turn down noisy third-party debug logs
logging.getLogger('spotipy').setLevel(logging.INFO)
logging.getLogger('urllib3').setLevel(logging.INFO)
# Logs on error for 404s (which we don't want to see)
logging.getLogger('spotipy.client').setLevel(logging.FATAL)

logger = logging.getLogger(__name__)

COMMANDS = ["copy", "commands", "daily", "init-config", "list-playlists", "randomize", "randomize-all", "subscriptions"]

COMMAND_DESCRIPTION = f"""
{'COMMAND NAME':<16} DESCRIPTION
{'commands':<16} Prints this command list.
{'copy':<16} Copies a playlist into a new playlist. You may optionally specify a destination playlist name.
{'daily':<16} Add "Daily Mix" entries to the end of the target playlist, filtering for excluded entries.
{'init-config':<16} Initializes the configuration file. Uses the --config_file location as the target. Will not overwrite.
{'list-playlists':<16} Prints a table describing your playlists.
{'randomize':<16} Randomizes the playlists with the given names, IDs, or in the given collections.
{'randomize-all':<16} Randomizes all playlists owned by the current user.
{'subscriptions':<16} Add new tracks from configured playlists to the target playlist, filtering for excluded entries.
"""

LOG_FORMAT = "%(levelname)s (%(name)s): %(message)s"
# LOG_FORMAT = "%(message)s"


def print_commands():
    """Prints available commands."""
    print(COMMAND_DESCRIPTION)


def append_daily_mix(config: Dict[str, Any], args: argparse.Namespace):
    """ Appends the daily mix to the target playlist.

    :param config: The configuration dictionary.
    :param args: The arguments namespace.
    """
    sp = get_spotify_handle(config)

    playlists = Playlists(sp, config.get("playlists"))
    playlists.append_daily_mix(args.randomize, args.target)


def append_recent_subscriptions(config: Dict[str, Any], args: argparse.Namespace):
    sp = get_spotify_handle(config)

    playlists = Playlists(sp, config.get("subscriptions"))
    playlist_set_filter = ValueFilter.from_nested(
        includes=args.include_playlist_sets, excludes=args.exclude_playlist_sets)
    playlists.append_recent_subscriptions(args.randomize, args.target, playlist_set_filter)


def randomize_lists(config: Dict[str, Any], args: argparse.Namespace):
    sp = get_spotify_handle(config)
    playlists = Playlists(sp, config.get("playlists"))

    if args.arguments:
        results = playlists.randomize_playlists(args.arguments)
    else:
        logger.warning("No playlists specified; nothing to randomize")
        return 5

    for item, result in results.items():
        print(f"{item}: {result.label}")


def randomize_all_lists(config: Dict[str, Any], args: argparse.Namespace):
    sp = get_spotify_handle(config)
    playlists = Playlists(sp, config.get("playlists"))

    results = playlists.randomize_owned_playlists()
    if results:
        logger.debug(f"Randomizing {len(results)} playlists")
    else:
        logger.warning("No user-owned playlists found to randomize")
        return 5

    for item, result in results.items():
        print(f"{item}: {result.label}")


def copy_list(config: Dict[str, Any], args: argparse.Namespace):
    arguments = args.arguments
    if arguments:
        sp = get_spotify_handle(config)
        playlists = Playlists(sp, config.get("playlists"))
        result, copy_name = playlists.copy_list(arguments, args.randomize)

        if PlaylistResult.SUCCESS == result:
            print(f"Successfully copied {arguments[0]} to {copy_name}")
        else:
            logger.warning(f"Problems copying {arguments[0]}")

    else:
        logger.warning("No playlist specified; nothing to copy")
        return 6


def list_playlists(config: Dict[str, Any], args: argparse.Namespace):
    sp = get_spotify_handle(config)
    playlists = Playlists(sp, config.get("playlists"))

    try:
        all_playlists = playlists.list_all_playlists(
            filters=args.filters, sort_fields=args.sort_fields
        )
    except Exception as e:
        logger.warning(f"Problems listing playlists: {e}")
        return 1

    print(f"{'PLAYLIST NAME':<32} {'SIZE':<5} {'ID':<24} {'OWNER':<16} DESCRIPTION")
    for playlist_row in all_playlists:
        print(
            f"""{truncate_long_value(playlist_row[FieldName.PLAYLIST_NAME], 32):<32} \
{playlist_row[FieldName.SIZE]:<5} \
{playlist_row[FieldName.SPOTIFY_ID]:<24} \
{truncate_long_value(playlist_row[FieldName.OWNER], 16):<16} \
{truncate_long_value(playlist_row[FieldName.PLAYLIST_DESCRIPTION], 75)}"""
        )


# The basic config structure for the CLI
initial_config = {
    "spotify": {
        "client_id": "NO_SPOTIFY_CLIENT_ID",
        "client_secret": "NO_SPOTIFY_CLIENT_ID"
    },
    "playlists": {
        "daily_mix_target": "Now",
        "daily_mix_prefix": "Daily Mix",
        "daily_mix_excludes": "Overplayed"
    },
    "subscriptions": {
        "subscriptions_target": "NewSubscriptions",
        "max_age": "3 days",
        "playlists": {
            "noisy": ["2uiYiQFpynkWCpIXcBGir9", "37i9dQZF1DX8gDIpdqp1XJ", "3JEvwuKbVKoggEA75gWqET"],
            "jazz": ["37i9dQZF1DX7YCknf2jT6s", "37i9dQZF1DWUb0uBnlJuTi", "4xRrCdkn4r5lrDOElek5oC",
                     "2puFFdGTID0iJdQtjLvhal"]
        }
    }
}


def playlist_sets(value: str | None) -> list[str]:
    """Parses the playlist groups from the command line."""
    if not value:
        return [""]
    return [clean.strip() for clean in re.split("[,|]", value)]


def init_config(args: argparse.Namespace):
    config_file = args.config_file

    if config_file.exists():
        logger.warning(f"Config file at {config_file} already exists. Not overwriting.")
        return 10

    config_file.parent.mkdir(parents=True, exist_ok=True)

    with open(config_file, mode="wb") as fp:
        tomli_w.dump(initial_config, fp)


class CommandLookup(BaseLookup):

    def _init_lookup(self):
        lookup = pygtrie.CharTrie()
        lookup["cop"] = "copy"
        lookup["com"] = "commands"
        lookup["d"] = "daily"
        lookup["i"] = "init-config"
        lookup["l"] = "list-playlists"
        lookup["randomize-a"] = "randomize-all"
        lookup["r"] = "randomize"
        lookup["s"] = "subscriptions"
        return lookup

def parse_cmdline(argv: List):
    """
    Returns the parsed argument list and return code.
    `argv` is a list of arguments, or `None` for ``sys.argv[1:]``.
    """
    if argv is None:
        argv = sys.argv[1:]

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config_file",
                        help=f"The location of the config file (default: {DEFAULT_CONFIG_FILE})",
                        default=DEFAULT_CONFIG_FILE, type=Path)
    parser.add_argument("-f", "--filters", help="Filters to apply to the list")
    parser.add_argument("-e", "--exclude_playlist_sets", type=playlist_sets,
                        action="append", help="The playlist sets to exclude. Takes precedence over include")
    parser.add_argument("-i", "--include_playlist_sets", type=playlist_sets,
                        action="append", help="The playlist sets to include. Includes all if not specified")
    parser.add_argument("-r", "--randomize", help="Randomize the target list", action='store_true')
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument("-s", "--sort_fields", help="The fields to sort against, applied in order")
    parser.add_argument("-t", "--target",
                        help="Specify the target name of the operation (overrides any default value)")
    parser.add_argument("command", metavar="COMMAND",
                        help=f"The command to run (one of {','.join(COMMANDS)})")
    parser.add_argument("arguments", metavar='ARGUMENTS', nargs='*',
                        help="the arguments to the command")
    # noinspection PyTypeChecker
    parser.add_argument('-log',
                        '--loglevel',
                        default='info',
                        type=str.upper,
                        choices=logging._nameToLevel.keys(),
                        help='Provide logging level. Example: --loglevel debug, default is info')
    args = None
    try:
        args = parser.parse_args(argv)
    except OSError as e:
        logger.warning("Problems reading file:", e)
        parser.print_help()
        return args, 2

    return args, 0


def main(argv=None):
    args, ret = parse_cmdline(argv)
    if ret != 0:
        return ret
    logging.basicConfig(format=LOG_FORMAT, level=args.loglevel)

    config = get_config(args.config_file)

    command = CommandLookup().find(args.command.lower())

    if command == "commands":
        return print_commands()
    elif command == "copy":
        return copy_list(config, args)
    elif command == "daily":
        return append_daily_mix(config, args)
    elif command == "init-config":
        return init_config(args)
    elif command == "list-playlists":
        return list_playlists(config, args)
    elif command == "randomize":
        return randomize_lists(config, args)
    elif command == "randomize-all":
        return randomize_all_lists(config, args)
    elif command == "subscriptions":
        return append_recent_subscriptions(config, args)
    else:
        print(
            f"Invalid command '{args.command}'.  Valid commands: {','.join(COMMANDS)}"
        )
        return 1


if __name__ == "__main__":
    status = main()
    sys.exit(status)



import datetime
import itertools
import logging
from abc import ABC, abstractmethod
from itertools import islice
from pathlib import Path
from typing import Iterable, Dict, Any

import spotipy
import tomli
from appdirs import user_config_dir, user_cache_dir
from pygtrie import CharTrie

from spotipy import Spotify

DEFAULT_CONFIG_DIR = user_config_dir("spotcrates")
DEFAULT_CONFIG_FILE = Path(DEFAULT_CONFIG_DIR, "spotcrates_config.toml")
DEFAULT_CACHE_DIR = user_cache_dir("spotcrates")
DEFAULT_AUTH_CACHE_FILE = Path(DEFAULT_CACHE_DIR, "spotcrates_auth_cache")
DEFAULT_AUTH_SCOPES = ["playlist-modify-private", "playlist-read-private"]
DEFAULT_REDIRECT_URI = 'http://127.0.0.1:5000/'
DEFAULT_TARGET = "default_target"
ISO_8601_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
ZERO_TIMESTAMP = datetime.datetime.strptime(
    '1970-01-01T00:00:00Z', ISO_8601_TIMESTAMP_FORMAT
)

class NotFoundException(Exception):
    pass


def batched(iterable: Iterable, n: int):
    """Batch data into lists of length n. The last batch may be shorter."""
    it = iter(iterable)
    while True:
        batch = list(islice(it, n))
        if not batch:
            return
        yield batch


def get_all_items(spotify: Spotify, first_page: Dict[str, Any]):
    """Collects the 'items' contents from every page in the given result set."""
    all_items = []

    all_items.extend(first_page["items"])

    try:
        next_page = spotify.next(first_page)
        while next_page:
            all_items.extend(next_page["items"])
            next_page = spotify.next(next_page)
    except TimeoutError:
        logging.info("Timeout error encountered while paging Spotify items")
    except Exception as e:
        logging.warning(f"Problems paging given Spotify items list: {e}", exc_info=True)

    return [item for item in all_items if item is not None]


def truncate_long_value(full_value: str | None, length: int, trim_tail: bool = True) -> str:
    """Returns the given value truncated from the start of the value so that it is at most the given length.

    :param full_value: The value to trim.
    :param length: The maximum length of the returned value.
    :param trim_tail: Whether to trim from the head or tail of the string.
    :return: The value trimmed from the start of the string to be at most the given length.
    """
    if not full_value:
        return full_value

    if len(full_value) > length:
        if trim_tail:
            return full_value[:length]
        else:
            return full_value[-length:]
    return full_value


class BaseLookup(ABC):
    def __init__(self):
        """Uses a trie to find the longest matching prefix for the given lookup value"""
        self.logger = logging.getLogger(__name__)
        self.lookup = self._init_lookup()

    def find(self, lookup_val):
        if not lookup_val:
            raise NotFoundException(f"Blank/null lookup value {lookup_val}")

        found_command = self.lookup.longest_prefix(lookup_val)

        if found_command:
            self.logger.debug(
                "Got %s (%s) for %s", found_command.value, found_command.key, lookup_val
            )
            return found_command.value
        else:
            raise NotFoundException(f"No value for {lookup_val}")

    @abstractmethod
    def _init_lookup(self) -> CharTrie:
        pass


def get_spotify_handle(config: Dict[str, Dict[str, Any]]):
    spotify_cfg = config.get("spotify")
    if not spotify_cfg:
        raise Exception("No Spotify config defined")
    auth_scopes = spotify_cfg.get("auth_scopes", DEFAULT_AUTH_SCOPES)
    redirect_uri = spotify_cfg.get("redirect_uri", DEFAULT_REDIRECT_URI)
    cache_path = prepare_auth_cache_loc(spotify_cfg)
    cache_handler = spotipy.cache_handler.CacheFileHandler(cache_path=cache_path)
    if spotify_cfg:
        auth_manager = spotipy.oauth2.SpotifyOAuth(
            client_id=spotify_cfg.get("client_id"),
            client_secret=spotify_cfg.get("client_secret"),
            redirect_uri=redirect_uri,
            cache_handler=cache_handler,
            scope=auth_scopes,
        )
    else:
        auth_manager = spotipy.oauth2.SpotifyOAuth(
            cache_handler=cache_handler, scope=auth_scopes
        )
    return spotipy.Spotify(auth_manager=auth_manager)


def prepare_auth_cache_loc(config: Dict[str, Any]):
    auth_cache_file = Path(config.get("auth_cache", DEFAULT_AUTH_CACHE_FILE))
    if not auth_cache_file.exists():
        auth_cache_file.parent.mkdir(parents=True, exist_ok=True)
    return auth_cache_file


def get_config(config_file: Path):
    """Loads the specified TOML-formatted config file or returns an empty dict"""
    if config_file.exists():
        with open(config_file, mode="rb") as fp:
            return tomli.load(fp)
    else:
        logging.debug(f"Config file '{config_file}' does not exist")
        return {}


class ValueFilter:
    def __init__(self, includes: list[str] | None = None, excludes: list[str] | None = None):
        self._includes = includes
        self._excludes = excludes

    def include(self, value: str) -> bool:
        if self._excludes and value in self._excludes:
            return False
        return not self._includes or value in self._includes

    def exclude(self, value: str) -> bool:
        return not self.include(value)

    @classmethod
    def from_nested(cls, includes: list[list[str]] | None = None, excludes: list[list[str]] | None = None):
        flat_includes = list(itertools.chain.from_iterable(includes)) if includes else includes
        flat_excludes = list(itertools.chain.from_iterable(excludes)) if excludes else excludes
        return cls(flat_includes, flat_excludes)


import datetime
import logging
import random
from contextlib import suppress
from enum import Enum
from typing import List, Set, Dict, Tuple, Any, Iterable

from durations_nlp import Duration
from spotipy import Spotify, SpotifyException

from spotcrates.common import batched, get_all_items, ISO_8601_TIMESTAMP_FORMAT, ZERO_TIMESTAMP, ValueFilter
from spotcrates.filters import FieldName, filter_list, sort_list

config_defaults = {
    "daily_mix_prefix": "Daily Mix",
    "daily_mix_target": "Now",
    "daily_mix_exclude_prefix": "Overplayed",
    "subscriptions_target": "NewSubscriptions",
    "max_age": "3 days",
    "include_zero_timestamps": True,
    "playlists": {},
    "oldest_timestamp": None,
}


class PlaylistException(Exception):
    pass


class PlaylistConfigException(PlaylistException):
    pass


class PlaylistNamingException(PlaylistException):
    pass


# True, False, File not found
class PlaylistResult(Enum):
    """Represents the result of a playlist operation."""

    def __init__(self, label):
        self.label = label

    SUCCESS = "Success"
    FAILURE = "Failure"
    NOT_FOUND = "Not Found"


class Playlists:

    def __init__(self, spotify: Spotify, config: Dict | None = None):
        """Creates an instance of the playlist manipulation class.

        :param spotify: A handle for the initialized SpotiPy client.
        :param config: The configuration for the playlists class.
        """
        self.spotify = spotify
        self.logger = logging.getLogger(__name__)

        self.config = self._process_config(config)

    def get_all_playlists(self) -> List[Dict]:
        return get_all_items(self.spotify, self.spotify.current_user_playlists())

    def list_all_playlists(self, sort_fields=None, filters=None) -> List[Dict]:
        playlist_entries = []
        for playlist in self.get_all_playlists():
            playlist_entries.append(
                {
                    FieldName.SPOTIFY_ID: playlist["id"],
                    FieldName.PLAYLIST_NAME: playlist["name"],
                    FieldName.SIZE: playlist["tracks"]["total"],
                    FieldName.OWNER: playlist["owner"]["id"],
                    FieldName.PLAYLIST_DESCRIPTION: playlist["description"],
                }
            )

        if not sort_fields and not filters:
            return playlist_entries

        processed_entries = playlist_entries
        if filters:
            processed_entries = filter_list(processed_entries, filters)

        if sort_fields:
            processed_entries = sort_list(processed_entries, sort_fields)

        return processed_entries

    def append_daily_mix(self, randomize: bool, target_name: str):
        """Combines all of the "daily mix" playlists for the account and removes any
        tracks found in the "exclude" playlists along with any that are already in
        the given target list. Optionally randomizes the remainder before appending to
        (or creating) the target list.

        :param randomize: Whether to randomize the "daily mix" tracks before appending to
        the target list.
        :param target_name: The name of the list to append to or create. The configured
        list at 'daily_mix_target' is used if no name is provided here.
        """
        dailies = []
        target_list = None
        exclude_lists = []
        daily_mix_prefix = self.config.get("daily_mix_prefix")
        if target_name:
            daily_mix_target = target_name
        else:
            daily_mix_target = self.config.get("daily_mix_target")

        daily_mix_exclude_prefix = self.config.get("daily_mix_exclude_prefix")
        for playlist in self.get_all_playlists():
            list_name = playlist.get("name")
            if list_name:
                if list_name.startswith(daily_mix_prefix):
                    dailies.append(playlist)
                elif list_name == daily_mix_target:
                    target_list = playlist
                elif list_name.startswith(daily_mix_exclude_prefix):
                    exclude_lists.append(playlist)

        # TODO: Optionally create if it doesn't exist
        if not target_list:
            me = self.spotify.me()

            target_list = self.spotify.user_playlist_create(
                me["id"], daily_mix_target, public=False
            )

        # user_playlist_add_tracks(user, playlist_id, tracks, position=None)
        if not dailies:
            self.logger.warning(
                f"No daily mixes found with the prefix '{daily_mix_prefix}'"
            )
            return

        exclude_ids = self._get_excludes(exclude_lists, target_list)

        add_tracks, orig_daily_count = self._fetch_daily_tracks(dailies, exclude_ids)

        self.logger.info(
            f"{len(add_tracks)} to add from an original count of {orig_daily_count}"
        )
        if add_tracks:
            if randomize:
                random.shuffle(add_tracks)
            self._add_tracks_to_playlist(target_list, add_tracks)
        else:
            self.logger.warning("No daily songs to add")

    def randomize_playlist(self, playlist: Dict[str, Any]) -> PlaylistResult:
        """Randomizes the tracks in the given playlist.

        :param playlist: The name of the playlist to randomize.
        :return: The result of the randomization attempt.
        """
        try:
            playlist_tracks = self._get_playlist_id_tracks(playlist["id"])
            if not playlist_tracks:
                self.logger.warning(f"No tracks found for playlist '{playlist['name']}'")
                return PlaylistResult.FAILURE
            random.shuffle(playlist_tracks)
            self._add_tracks_to_playlist(playlist, playlist_tracks, replace_playlist=True)
            return PlaylistResult.SUCCESS
        except Exception:
            self.logger.warning(f"Problems randomizing playlist '{playlist['name']}'", exc_info=True)
            return PlaylistResult.FAILURE

    def append_recent_subscriptions(
            self, randomize: bool, target_name: str, playlist_set_filter: ValueFilter| None = None):
        """Adds new tracks from configured lists to the target list. Newness is determined
        based on the configured maximum age, which is 3 days by default.

        :param randomize: Whether to randomize the new tracks before appending them.
        :param target_name: The target list to append to. The configured target list at
          'subscriptions_target' is used if no name is provided here.
        :param playlist_set_filter: A filter to apply to the playlist sets.
        """
        target_list = None
        exclude_lists = []
        if target_name:
            subscriptions_target = target_name
        else:
            subscriptions_target = self.config.get("subscriptions_target")
        exclude_prefix = self.config.get("daily_mix_exclude_prefix")

        if playlist_set_filter is None:
            playlist_set_filter = ValueFilter()

        for playlist in self.get_all_playlists():
            list_name = playlist["name"]
            if list_name:
                if list_name == subscriptions_target:
                    target_list = playlist
                elif list_name.startswith(exclude_prefix):
                    exclude_lists.append(playlist)

        oldest_timestamp = self._get_oldest_timestamp()

        # TODO: Optionally create if it doesn't exist
        if not target_list:
            me = self.spotify.me()

            target_list = self.spotify.user_playlist_create(
                me["id"], subscriptions_target, public=False
            )

        excludes = self._get_excludes(exclude_lists, target_list)

        include_zero_timestamps = self.config.get("include_zero_timestamps", False)
        playlist_ids = self._get_subscription_playlist_ids(
            oldest_timestamp, excludes, include_zero_timestamps, playlist_set_filter)

        self.logger.info(f"{len(playlist_ids)} subscription tracks to add")

        if randomize:
            playlist_ids = list(playlist_ids)
            random.shuffle(playlist_ids)

        for id_batch in batched(playlist_ids, 100):
            self.logger.debug(f"Batch size: {len(id_batch)}")
            self.spotify.playlist_add_items(target_list["id"], id_batch)

    def _get_oldest_timestamp(self):
        cfg_oldest_timestamp = self.config.get("oldest_timestamp")

        if cfg_oldest_timestamp:
            try:
                return datetime.datetime.strptime(
                    cfg_oldest_timestamp, ISO_8601_TIMESTAMP_FORMAT
                )
            except Exception:
                self.logger.warning(f"Could not parse oldest_timestamp value {cfg_oldest_timestamp}",
                                    exc_info=True)

        max_age = "NO_MAX_AGE"
        try:
            max_age = self.config.get("max_age", "NO_MAX_AGE")

            return datetime.datetime.now() - datetime.timedelta(
                seconds=Duration(max_age).to_seconds()
            )
        except Exception as e:
            raise PlaylistConfigException(f"Could not parse track age {max_age}", e)

    def randomize_playlists(self, playlists: List[str]) -> Dict[str, PlaylistResult]:
        """Replaces the tracks in the target lists with a randomized version of the same tracks.

        :param playlists: A list of playlist names and/or IDs.
        :return: The results of the randomizing.
        """

        lower_playlists = [playlist.lower() for playlist in playlists]

        results = {}
        for playlist in self.get_all_playlists():
            list_name = playlist["name"]
            if list_name and list_name.lower() in lower_playlists:
                results[list_name] = self.randomize_playlist(playlist)
            else:
                list_id = playlist['id']
                if list_id in playlists:
                    results[list_id] = self.randomize_playlist(playlist)

        found_targets = results.keys()
        missing_targets = [missing for missing in playlists if missing not in found_targets]
        for missing_target in missing_targets:
            results[missing_target] = PlaylistResult.NOT_FOUND

        return results

    def copy_list(self, arguments: List[str], randomize: bool) -> Tuple[PlaylistResult, str | None]:
        try:
            source_name = arguments[0]
            if len(arguments) < 2:
                dest_name = self._create_unique_dest_name(source_name)
            else:
                dest_name = arguments[1]

            # TODO: make public flag settable
            new_playlist = self.spotify.user_playlist_create(self.spotify.me()['id'], dest_name, public=False)
            tracks_to_copy = self._get_playlist_name_tracks(source_name)

            if randomize:
                random.shuffle(tracks_to_copy)

            self._add_tracks_to_playlist(new_playlist, tracks_to_copy)
            return PlaylistResult.SUCCESS, dest_name
        except Exception:
            self.logger.warning("Problems copying list", exc_info=True)
            return PlaylistResult.FAILURE, None

    def randomize_owned_playlists(self) -> Dict[str, PlaylistResult]:
        """Randomizes all playlists owned by the current user.

        Returns:
            Dict mapping playlist names to their randomization results
        """
        results = {}
        current_user_id = self.spotify.current_user()["id"]

        for playlist in self.get_all_playlists():
            if playlist["owner"]["id"] == current_user_id:
                results[playlist["name"]] = self.randomize_playlist(playlist)

        return results

    # ===Internal Methods=== #

    def _fetch_daily_tracks(self, dailies: List, exclude_ids: Iterable[str]):
        add_tracks = []
        orig_daily_count = 0
        for daily in dailies:
            daily_items = self._get_playlist_id_tracks(daily["id"])
            orig_daily_count += len(daily_items)
            add_tracks.extend(
                [
                    daily_item
                    for daily_item in daily_items
                    if daily_item["track"]["id"] not in exclude_ids
                ]
            )
        return add_tracks, orig_daily_count

    def _get_subscription_playlist_ids(self,
                                       oldest_timestamp: datetime,
                                       excluded_ids: Iterable[str],
                                       include_zero_timestamps: bool,
                                       playlist_set_filter: ValueFilter) -> Set[str]:
        target_playlist_ids: Set[str] = set()
        subscription_playlists = self.config.get("playlists")
        if not subscription_playlists:
            self.logger.warning("No subscription playlists defined")
            return target_playlist_ids

        for playlist_set, playlist_ids in subscription_playlists.items():
            self.logger.debug(f"Processing subscription set '{playlist_set}'")
            if playlist_set_filter.exclude(playlist_set):
                self.logger.debug(f"Skipping set '{playlist_set}'")
                continue
            set_playlist_ids = set()
            for track in self._get_playlist_id_tracks(*playlist_ids):
                iso_added = track.get("added_at")
                if iso_added:
                    try:
                        track_timestamp = datetime.datetime.strptime(
                            iso_added, ISO_8601_TIMESTAMP_FORMAT
                        )
                        if self._include_for_added_at(oldest_timestamp, track_timestamp, include_zero_timestamps):
                            track_id = track.get("track", {}).get("id")
                            if track_id and track_id not in excluded_ids:
                                set_playlist_ids.add(track_id)
                    except Exception:
                        self.logger.warning(
                            f"Could not parse timestamp {iso_added}. Skipping track.",
                            exc_info=True,
                        )
                else:
                    logging.debug("No 'added_at' field for track. Skipping.")
            self.logger.debug(
                f"Found {len(set_playlist_ids)} newer than {oldest_timestamp} "
                f"in playlist set '{playlist_set}'"
            )
            target_playlist_ids.update(set_playlist_ids)

        self.logger.debug(
            f"Found a total of {len(target_playlist_ids)} newer than {oldest_timestamp}"
        )
        return target_playlist_ids

    @staticmethod
    def _include_for_added_at(oldest_timestamp: datetime,
                              track_timestamp: datetime,
                              include_zero_timestamps: bool) -> bool:
        if include_zero_timestamps:
            return track_timestamp >= oldest_timestamp or track_timestamp == ZERO_TIMESTAMP
        else:
            return track_timestamp >= oldest_timestamp

    def _get_playlist_track_ids(self, *args: str) -> Set[str]:
        track_ids: Set[str] = set()
        for playlist_id in args:
            playlist_items = self._filter_for_tracks(playlist_id)

            track_ids.update(
                {playlist_item.get("track", {}).get("id") for playlist_item in playlist_items}
            )

        # Remove None in case any IDs failed to resolve
        with suppress(KeyError):
            # noinspection PyTypeChecker
            track_ids.remove(None)  # type: ignore
        return track_ids

    def _get_playlist_names_to_ids(self, lower=False) -> Dict[str, str]:
        name_ids = {}
        for playlist in self.get_all_playlists():
            if lower:
                name_ids[playlist['name'].lower()] = playlist['id']
            else:
                name_ids[playlist['name']] = playlist['id']

        return name_ids

    def _get_playlist_name_tracks(self, *playlist_names: str) -> List[dict]:
        name_id_map = self._get_playlist_names_to_ids(lower=True)
        playlist_ids = []
        for lower_name in [name.lower() for name in playlist_names]:
            playlist_id = name_id_map.get(lower_name)
            if playlist_id:
                playlist_ids.append(playlist_id)
            else:
                self.logger.warning(f"No playlist found for name '{lower_name}'")
        return self._get_playlist_id_tracks(*playlist_ids)

    def _get_playlist_id_tracks(self, *playlist_ids: str) -> List[dict]:
        tracks = []
        for playlist_id in playlist_ids:
            filtered_tracks = self._filter_for_tracks(playlist_id)

            tracks.extend(filtered_tracks)

        return tracks

    def _filter_for_tracks(self, playlist_id: str) -> List[dict]:
        try:
            all_tracks = get_all_items(self.spotify, self.spotify.playlist_items(playlist_id))
        except SpotifyException as e:
            if e.http_status == 404:
                self.logger.info(f"Playlist '{playlist_id}' not found")
                return []
            else:
                self.logger.warning(f"Spotify problems getting tracks for playlist '{playlist_id}': {e}")
                return []
        except Exception as e:
            self.logger.warning(f"Problems getting tracks for playlist '{playlist_id}': {e}")
            return []
        filtered_tracks = []
        for cur_track in all_tracks:
            if cur_track.get('track') and cur_track['track'].get('id'):
                filtered_tracks.append(cur_track)
        return filtered_tracks

    def _add_tracks_to_playlist(self, target_list: Dict[str, Any], add_tracks: List, replace_playlist=False):
        track_ids = {add_song["track"]["id"] for add_song in add_tracks}

        first_batch = True
        for id_batch in batched(track_ids, 100):
            if replace_playlist and first_batch:
                self.spotify.playlist_replace_items(target_list["id"], id_batch)
            else:
                self.spotify.playlist_add_items(target_list["id"], id_batch)
            self.logger.debug(f"Batch size: {len(id_batch)}")
            first_batch = False

    def _get_excludes(self, exclude_lists: List[Dict], target_list: Dict):
        exclude_ids = self._get_playlist_track_ids(target_list["id"])
        for exclude_list in exclude_lists:
            exclude_ids.update(self._get_playlist_track_ids(exclude_list["id"]))
        return exclude_ids

    @staticmethod
    def _process_config(config: Dict | None) -> Dict:
        processed_config = {}

        source_config = {}
        if config:
            source_config = config

        for key, default_value in config_defaults.items():
            processed_config[key] = source_config.get(key, default_value)

        return processed_config

    def _get_playlist_names(self, lower=False) -> List[str]:
        if lower:
            return [playlist["name"].lower() for playlist in self.get_all_playlists()]
        else:
            return [playlist["name"] for playlist in self.get_all_playlists()]

    def _create_unique_dest_name(self, source_name: str):
        existing_lists = self._get_playlist_names(lower=True)
        count = 1

        dest_name = f"{source_name}-{count:02d}"

        if dest_name.lower() not in existing_lists:
            return dest_name

        max_count = 99

        while count <= max_count:
            count += 1
            dest_name = f"{source_name}-{count:02d}"

            if dest_name.lower() not in existing_lists:
                return dest_name

        raise PlaylistNamingException(f"Unable to find a unique playlist name (gave up at {dest_name})")
